/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file audio.h
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 25.02.2006
 * 
 * Audio module. This handles the task of reading audio files and playing
 * the audio output. It uses a MP3-decoder and libao, and it can easilly be
 * extended with more decoders.
 */


#ifndef AUDIO_H_
#define AUDIO_H_

#include <mad.h>		/* mad_timer_t */
#include <ao/ao.h>		/* ao_sample_format */
#include <sys/stat.h>	/* file stat */
#include <pthread.h>

#include "daisylibao.h"
#include "libdaisy.h"


/* ************************************************	*
 * 	Global type definitions and datastructures		*
 * ************************************************	*/

/**
 * Points to a struct_audio_data_t.
 * This gives the user a possibility to refer to a struct_audio_data_t,
 * but not access to it's internal data.
 */
typedef void * audio_data_t;

/**
 * Type to identify different audio file formats.
 */
typedef enum AUDIO_FILE_TYPE 
{
	AUDIO_FILE_WAV,				/* PCM WAVE file */
	AUDIO_FILE_MP3,				/* MP3 file. uses mp3.h */
	AUDIO_FILE_AAC,				/* AAC file. uses aac.h */
	AUDIO_FILE_UNKNOWN			/* Unknown filetype */
} audio_file_t;

/**
 * Audio engine states.
 * These are the different states the audio-engine can be in.
 */
typedef enum {
	AUDIO_STATE_PLAY,
	AUDIO_STATE_PAUSE,
	AUDIO_STATE_STOP,
	AUDIO_STATE_NOOP
} audio_state_t;

/**
 * This is the audio data.
 * All information related to the audio should be in this struct. This struct is passed along to almost all audio engine functions.
 */
typedef struct
{
	/* buffer */
	struct buffer
	{
		unsigned char const *buffer_start;
		unsigned long buffer_length;
	} buffer;

	char			*file_name;			/* the file_name of the file we have loaded */
	char			*prev_file_name;	/* the previous file_name. we check against this so that we won't have to reload already opened files */

	int				fd;					/* input file descriptor */
	void			*fdm;				/* pointer to file mmap */
	struct stat		stat;				/* file stats */

	audio_file_t	file_type;			/* the file_type (mp3, aac, wav) */
	audio_file_t	prev_file_type;		/* the previous file_type */

	ao_sample_format format;			/* the format of our output samples. we use this to initialize ao */

	/* mad_timer_t variables to keep track of progress, start and stop-times. */
	/* TODO: We'll have to use this for all decoders. Define our own struct in audio.h and use this instead? This should be fixed before additional decoders are implemented. */
	mad_timer_t		progress;
	mad_timer_t		start;
	mad_timer_t		stop;

	audio_state_t state;
	bool is_playing;
	
	bool first_run;	/* we use this to find out if a decoder iteration is our first or not. */
	
	/* thread data */
	pthread_mutex_t		mutex;			/* we need to prevent race conditions on shared variables */

	pthread_cond_t		finished_cond;
	pthread_cond_t		stopped_cond;
	pthread_cond_t		playing_cond;
	pthread_cond_t		unpause_cond;
	
	/* callback functions */
	void (*cb_playing_done) (daisyplayer_t d);	/* audio done callback function */
	bool do_callback_done;						/* set to true to run cb_playing_done once */
	void (*cb_error) (void *data, enum daisy_status, const char *);	/* error callback */
	void (*cb_progress) (void *data, long int progress_ms);	/* progress callback */
	
	daisyplayer_t *daisy;				/* this will need to be passed along when callback function cb_playing_done is called */

	struct audio_dither *mp3_dither;	/* Dither struct. This is used by the mp3.c dither routine */
	
	ao_device *device;					/* device is used by ao. */
	
	unsigned int magic;					/* magic number */
} struct_audio_data_t;



/* ************************************************	*
 * 			Global function declarations  		  	*
 * ************************************************	*/

/**
 * Plays a file from start to stop.
 * @param audio_data_t _data - the struct_audio_data_t that is passed to almost all functions in the audio module.
 * @param const char *fname - the filename.
 * @param const char *start - the start time as a string in a DAISY compliant format.
 * @param const char *stop - the stop time as a string in a DAISY compliant format.
 * @return the duration in milliseconds or -1 if fail.
 */
unsigned
long
int audio_play (audio_data_t _data, const char *fname, const char *start, const char *stop);

/**
 * Pause audio.
 * Sets the audio engine state to pause.
 * @param audio_data_t _data - the struct_audio_data_t that is passed to almost all functions in the audio module.
 * @return 1 on success, -1 on fail.
 */
int audio_pause (audio_data_t _data);

/**
 * Stops audio.
 * Sets the state to stop and blocks to make sure that audio has actually stopped.
 * @param audio_data_t _data - the struct_audio_data_t that is passed to almost all functions in the audio module.
 * @return 1 on success, -1 on fail.
 */
int audio_stop (audio_data_t _data);

/**
 * callback is called by the decoder loop.
 * @param audio_data_t _data - the struct_audio_data_t that is passed to almost all functions in the audio module.
 * @param - progress_ms the progress (so far) in the current segment. value is in milliseconds.
 * @return 0 to continue or -1 to stop decoding.
 */
int callback (audio_data_t _data, long int);
 
/**
 * Stops and frees audio engine resources.
 * This is called when the user wants us to terminate.
 * @param audio_data_t _data - the struct_audio_data_t that is passed to almost all functions in the audio module.
 */
void audio_terminate (audio_data_t _data);

/**
 * Initiate the audio engine.
 * This is called when the program wants us to start and prepare for audio playback
 * This should allocate the basic resources needed by the audio engine.
 * @param - daisy the daisy struct
 * @param - done the callback function that will be called once a segment have been played.
 * @param - error the callback function that will be called when errors occur.
 * @param - progress the callback function that will be called during playback.
 */
audio_data_t audio_initiate (daisyplayer_t daisy,
							 void (*done) (daisyplayer_t d),
							 void (*error) (void *data, enum daisy_status, const char *),
							 void (*progress) (void *data, long int progress_ms));

/**
 * Audio thread.
 * This is run as a thread from daisyplayer.c and keeps the audio-engine alive.
 * audio_initiate MUST be called before this function.
 * audio_terminate MUST be called once you are done using the audio engine.
 * @param audio_data_t _data - the struct_audio_data_t that is passed to almost all functions in the audio module.
 */
void *audio_thread (void *_data);

/**
 * Returns the current state of the audio engine.
 * @param audio_data_t _data - the struct_audio_data_t that is passed to almost all functions in the audio module.
 * @return the currant state of the audio engine
 */
audio_state_t audio_get_state (audio_data_t _data);

#endif /* AUDIO_H_ */
