/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file libdaisy.h
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 21.03.2006
 * 
 * A daisy dtb engine.
 */
 
 
#ifndef LIBDAISY_H_
#define LIBDAISY_H_

#include <pthread.h>

#ifdef __cplusplus
extern "C" {
#endif



/* ************************************************	*
 * 	Global type definitions and datastructures		*
 * ************************************************	*/
 
/**
 * This will need to be initialized and passed along to daisy functions.
 */
typedef void *daisyplayer_t;

/**
 * daisy_status defines the different sort of status messages libdaisy might return.
 */
enum daisy_status 
{
	DAISY_ERROR_UNKNOWN							= 0x0000,	/* unknown error */

	/* parser related errors */
	

	/* audio related errors */
	DAISY_ERROR_AUDIO_NOT_INITIALIZED			= 0x0200,	/* audio isn't (properly) initialized */
	DAISY_ERROR_AUDIO_CREATE_MMAP				= 0x0201,	/* could not read file into memmory */
	DAISY_ERROR_AUDIO_FSTAT						= 0x0202,	/* could not get file stats */
	DAISY_ERROR_AUDIO_OPEN						= 0x0203,	/* could not open file */
	DAISY_ERROR_AUDIO_FREE_MMAP					= 0x0204,	/* could not free file from memory */
	DAISY_ERROR_AUDIO_INITIATE_DATA				= 0x0205,	/* could not initiate data struct */
	DAISY_ERROR_AUDIO_NOT_PLAYING				= 0x0206,	/* state is not playing */
	DAISY_ERROR_AUDIO_NOT_STOPPED				= 0x0207,	/* state is not stopped */
	DAISY_ERROR_AUDIO_DATA_IS_NULL				= 0x0208,	/* data struct seems to be NULL */
	DAISY_ERROR_AUDIO_MALLOC					= 0x0209,	/* failure allocating memmory */
	DAISY_ERROR_AUDIO_PAUSED_WHILE_NOT_PLAYING	= 0x0210,	/* paused when not playing */
	DAISY_ERROR_AUDIO_STOPPED_WHILE_NOT_PLAYING	= 0x0211,	/* stopped when not playing */
	
	/* playback related errors */
	DAISY_ERROR_PLAYBACK_NO_TEXT_IN_SEGMENT		= 0x0301,	/* no text in segment */
	DAISY_ERROR_PLAYBACK_NO_AUDIO_IN_SEGMENT	= 0x0302,	/* no audio in segment */
	DAISY_ERROR_PLAYBACK_NO_DTB_LOADED			= 0x0303,	/* no daisy dtb loaded */
	DAISY_ERROR_PLAYBACK_SEEK_FAILED			= 0x0304,	/* seek failed */
	
	/* other errors */
	DAISY_ERROR_MISC_INIT_MUTEX					= 0x0400,	/* could not initialize mutex */
	
	/* playback related status messages */
	DAISY_END_OF_BOOK							= 0x1100	/* we're at the end of the book */
};

/**
 * daisy_seek_option defines the different seek operations supported by the engine.
 * @see daisy_seek
 */
enum daisy_seek_option
{
	DAISY_SEEK_PREV_CHAPTER = 1,
	DAISY_SEEK_PREV_PASSAGE = 2,
	DAISY_SEEK_NEXT_CHAPTER = 3,
	DAISY_SEEK_NEXT_PASSAGE = 4,
	DAISY_SEEK_TO_BEGINNING = 5
};

/**
 * daisy_bookinfo_option defines different values to get bookinfo
 * @see daisy_get_info
 */
enum daisy_bookinfo_option
{
	DAISY_BOOKINFO_TITLETEXT = 1,
	DAISY_BOOKINFO_TITLEIMAGE = 2,
	DAISY_BOOKINFO_TOTALTIME = 3
};

/**
 * daisy_chapter_info defines different sort of chapter information.
 * @see daisy_get_chapter_info
 */
enum daisy_chapter_info
{
	DAISY_CHAPTER_TITLE = 1,
	DAISY_CHAPTER_WEIGHT = 2	
}; 

/**
 * This struct needs to be initalized and passeed along with the daisy API functions.
 * Contains pointers to the callback funtions.
 */
typedef struct {
	void 		*data;	/* user-defined pointer to data. this will be availible in all callback functions */
	void		(*cb_daisy_audio_done)	(void *);	/* (data, progress_ms_t);	called when libdaisy are done playing an audio segment */
	void		(*cb_daisy_audio_next)	(void *, unsigned long int);	/* (data, void) called when daisy are playing audio */
	void		(*cb_daisy_text)		(void *, void *);	/* (data, passage);	called when daisy have parsed a text-passage */
	void		(*cb_daisy_id)			(void *, void *);	/* (data, id);	called when daisy have parsed an id */
	void		(*cb_daisy_error)		(void *, enum daisy_status, const char *daisy_status_msg);
	void		(*cb_daisy_progress)	(void *, long int);	/* (data, progress_ms) */
	
	void *audio_data;	/* data related to audio decoding and playback */
	struct DaisyData *daisy_book_data;
	unsigned int magic;	/* magic number that verifies if the struct is valid or not */
	
	pthread_mutex_t daisy_mutex;	/* mutex used in libdaisy.c */

} struct_daisyplayer_t;

/**
 * This struct can hold a daisy DTB playback position. Ment for jumping to smil (chapter) and/or node position (passage).
 * Works as a bookmark.
 */
 typedef struct {
 	int smilpos;
 	int nodepos;
 } daisy_position;



/* ************************************************	*
 * 			Global function declarations  		  	*
 * ************************************************	*/

/**
 * This functions initializes the daisy library.
 * It should be called before any atempt to use daisy functionality.
 * @param data is a void pointer. This data pointer will be availible in all callback functions.
 * @param void l_cb_daisy_audio_done - this function will be called when an audio segment is done playing.
 * @param void l_cb_daisy_audio_next - this function will be called when an new audio segment starts playing, supplying the duration of the segment in ms.
 * @param void l_cb_daisy_text - this function will be called when an new audio segment starts playing, supplying the text corresponding to the audio.
 * @param void l_cb_daisy_id - this function will be called when an new audio segment starts playing, supplying the id of the text passage in the xml file.
 * @param void l_cb_daisy_error - this function will be called when an engine error occures.
 * @param void l_cb_daisy_progress - this function will be called during playback, supplying the progress in ms.
 * @returns daisyplayer_t - the daisy data stuct which must be passed along with all the API functions.
 */
daisyplayer_t daisy_init (void *data,
						  void (*l_cb_daisy_audio_done)	(void *),			/* (data, progress_ms_t ), */
						  void (*l_cb_daisy_audio_next)	(void *, unsigned long int),/* (data, void ), */
						  void (*l_cb_daisy_text)		(void *, void *),			/* (data, passage ), */
						  void (*l_cb_daisy_id)			(void *, void *),			/* (data, id ) */
						  void (*l_cb_daisy_error)		(void *, enum daisy_status, const char *daisy_status_msg),
						  void (*l_cb_daisy_progress)	(void *, long int)			/* (data, progress_ms) */
);

/**
 * This functions terminates the daisy library.
 * It should be called when daisy are no longer needed.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 */
void daisy_term (daisyplayer_t daisy);

/**
 * Loades a new daisy book ready for playback
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @param path - the path to the daisy dtb (ncc.* | *.ncx)
 * @returns 1 if success and -1 if not
 */
int daisy_load (daisyplayer_t daisy, char *path);

/**
 * Function to start playback.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @returns 1 on success and -1 if not.
 */
int daisy_play (daisyplayer_t daisy);

/**
 * Function to seek during playback.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @param seek_option: an daisy_seek_option.
 * @returns 0 if end_of_book, 1 if success and -1 on error
 */
int daisy_seek (daisyplayer_t daisy, int seek_option);

/**
 * Function to retrieve the current playback position.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @returns a pointer to a daisy_position struct containing the smil and node positions, or NULL in
 * case of error. The struct must be deallocated by the caller.
 */
daisy_position *daisy_get_position (daisyplayer_t daisy);

/**
 * Function for seeking to a bookmark and continue playback from there.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @param position - a pointer to a stuct with the smil and node position to jump to.
 * @returns 1 on success and -1 in case of error.
 */
int daisy_goto_position (daisyplayer_t daisy, daisy_position *position);

/**
 * Function to stop playback.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @returns 1 on success. (and always succeeds)
 */
int daisy_stop (daisyplayer_t daisy);

/**
 * Function to pause playback.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @returns 1 on success, -1 on fail.
 */
int daisy_pause (daisyplayer_t daisy);

/**
 * Function to get bookinfo
 * @param daisy - the daisy struct which must be passed along with all the API functions.
 * @param value - an daisy_bookinfo_option value which states what bookinfo value to get.
 * @returns the string if found, NULL otherwise. The string must be deallocated by the caller.
 */
char *daisy_get_info (daisyplayer_t daisy, int value);

/**
 * Function to get the number of chapters in a loaded daisy DTB.
 * @param daisy - the daisy struct which must be passed along with all the API functions.
 * @returns the number of chapters, or -1 on error.
 */
int daisy_get_chapter_count (daisyplayer_t daisy);

/**
 * Function for getting the information about a given chapter.
 * @param daisy - the daisy struct which must be passed along with all the API functions.
 * @param num - the number of the chapter to query on (use daisy_chapter_count to get the
 * number of chapters available).
 * @param option - an daisy_chapter_info which states what information to retrieve.
 * @returns a pointer to a string containing the queryed information, or NULL in case of error.
 */
char *daisy_get_chapter_info (daisyplayer_t daisy, int num, int option);

#ifdef __cplusplus
}
#endif


#endif /*LIBDAISY_H_*/

